# Created on Tue Dec 23 21:15:19 2014

# Author: XiaoTao Wang

import os, time, gc
import numpy as np
import pandas as pd
from cooler.util import load_fasta, read_chromsizes

def sleep():
    for _ in range(3):
        time.sleep(0.1)
    gc.collect()
    for _ in range(3):
        time.sleep(0.1)

def cleanFile(filename):
    if os.path.exists(filename):
        os.remove(filename)
    
def cleanDirectory(dirName):
    for i in os.listdir(dirName):
        os.remove(os.path.join(dirName, i))

def chromsizes_from_fasta(genomeFolder, genomeName):

    from Bio import SeqIO

    genomepath = os.path.join(genomeFolder, '.'.join([genomeName, 'fa']))
    chromsizes = {}
    fasta_genome = SeqIO.parse(genomepath, 'fasta')
    for fasta in fasta_genome:
        chromsizes[fasta.id] = len(fasta.seq)
    
    outfile = os.path.join(genomeFolder, '.'.join([genomeName, 'chrom', 'sizes']))
    with open(outfile, 'w') as out:
        for c in sorted(chromsizes):
            line = [c, str(chromsizes[c])]
            out.write('\t'.join(line)+'\n')
    
    return outfile

def chromsizes_from_pairs(pairpath):

    from pairtools import _fileio, _headerops

    instream = _fileio.auto_open(pairpath, mode='r')
    header, _ = _headerops.get_header(instream)
    genomeName = 'Unknown'
    chromsizes = []
    for r in header:
        if r.startswith('#genome_assembly:'):
            genomeName = r.split(':')[1].strip()
        if r.startswith('#chromsize:'):
            pair = r.split(':')[1].strip().split()
            chromsizes.append(pair)
    
    folder = os.path.split(pairpath)[0]
    outpath = os.path.join(folder, '.'+genomeName+'.chrom.sizes') # invisible to users
    with open(outpath, 'w') as out:
        for line in chromsizes: # order unchanged
            out.write('\t'.join(line)+'\n')
    
    instream.close()
    
    return outpath, genomeName

def digest(fasta_records, enzyme):
    """
    Divide a genome into restriction fragments. Support Arima-HiC enzyme cocktail
    which digest chromatin at ^GATC and G^ANTC.

    Parameters
    ----------
    fasta_records : OrderedDict
        Dictionary of chromosome names to sequence records.

    enzyme: str
        Name of restriction enzyme.

    Returns
    -------
    Dataframe with columns: 'chrom', 'start', 'end'.

    """
    import Bio.Restriction as biorst
    import Bio.Seq as bioseq
    # http://biopython.org/DIST/docs/cookbook/Restriction.html#mozTocId447698
    chroms = fasta_records.keys()
    try:
        if enzyme.lower()=='arima':
            cocktail = biorst.RestrictionBatch(['MboI', 'HinfI'])
            cut_finder = cocktail.search
        else:
            cut_finder = getattr(biorst, enzyme).search
    except AttributeError:
        raise ValueError('Unknown enzyme name: {}'.format(enzyme))

    def _each(chrom):
        seq = bioseq.Seq(str(fasta_records[chrom]))
        tmp = cut_finder(seq)
        if type(tmp)==list:
            cut_sites = tmp
        elif type(tmp)==dict:
            cut_sites = []
            for e in tmp:
                cut_sites.extend(tmp[e])
            cut_sites.sort()
        cuts = np.r_[0, np.array(cut_sites) + 1, len(seq)].astype(int)
        n_frags = len(cuts) - 1

        frags = pd.DataFrame({
            'chrom': [chrom] * n_frags,
            'start': cuts[:-1],
            'end': cuts[1:]},
            columns=['chrom', 'start', 'end'])
        return frags

    return pd.concat(map(_each, chroms), axis=0, ignore_index=True)


# Created on Sun Sep 9 15:04:12 2018

# Author: XiaoTao Wang

import subprocess, os, pickle
from runHiC.utilities import sleep
import numpy as np
from copy import deepcopy
import time
from collections import defaultdict

def merge_pairs(pair_paths, outpath, tmpdir):

    if len(pair_paths)==1:
        # Just make a soft link to original .pairsam
        os.symlink(pair_paths[0], outpath)
    else:
        # runHiC doesn't provide interface for changing detailed parameters of
        # pairtools merge for simplicity
        merge_command = ['pairtools', 'merge', '-o', outpath, '--nproc', '8', '--memory', '2G',
                         '--max-nmerge', '8', '--tmpdir', tmpdir] + pair_paths

        subprocess.check_call(' '.join(merge_command), shell=True)

def collect_stats(pair_paths):

    stats_pool = {}
    for i, p in enumerate(pair_paths):
        stats_path = p.replace('.pairsam.gz', '.pstats.1')
        with open(stats_path, 'rb') as source:
            stats_pool[str(i)] = pickle.load(source)['pseudo']
    
    merge_stats(stats_pool, list(stats_pool.keys()), 'pseudo')
    stats_pool = {'pseudo': stats_pool['pseudo']}

    return stats_pool


def stats_pairs(inpath, refkey, matchpre=[]):
    
    stat_command = ['pairtools', 'stats', inpath]
    pipe = subprocess.Popen(stat_command, stdout=subprocess.PIPE)
    inStream = pipe.stdout
    stats = defaultdict(int)
    for line in inStream:
        parse = line.decode().rstrip().split()
        key, value = parse[0], parse[1]
        if key in refkey:
            stats[refkey[key]] = int(value)
        for pre in matchpre:
            if key.startswith(pre):
                stats[key] = int(value)
    
    pipe.communicate()
    
    return stats

def stats_samfrag(samfrag_pairs, sample_size=100000):

    from pairtools import _fileio, _pairsam_format, _headerops

    instream = _fileio.auto_open(samfrag_pairs, mode='r')
    _, body_stream = _headerops.get_header(instream)

    stats = defaultdict(int)
    libsize = []
    for line in body_stream:
        cols = line.rstrip().split(_pairsam_format.PAIRSAM_SEP)
        pos1 = int(cols[_pairsam_format.COL_P1])
        strand1 = cols[_pairsam_format.COL_S1]
        pos2 = int(cols[_pairsam_format.COL_P2])
        strand2 = cols[_pairsam_format.COL_S2]
        #fragstart, fragend = int(cols[-2]), int(cols[-1]) # The index may change in the future
        stats['120_SameFragmentReads'] += 1
        if (strand1=='+') and (strand2=='-'): # dangling reads
            stats['124_DanglingReads'] += 1
            libsize.append(pos2-pos1)
        elif (strand1=='-') and (strand2=='+'): # self ligation
            stats['122_SelfLigationReads'] += 1
        else:
            stats['126_UnknownMechanism'] += 1
    
    instream.close()
    
    os.remove(samfrag_pairs)

    libsize = np.r_[libsize]
    np.random.shuffle(libsize)
    libsize = libsize[:sample_size]
    
    return stats, libsize

def create_frag(genomepath, chromsizes_file, enzyme, tmpdir):

    _, fastaName = os.path.split(genomepath)
    genomeName = fastaName.split('.')[0]
    prefix = '.'.join([genomeName, 'frags', enzyme])
    outbed = os.path.join(tmpdir, '.'.join([prefix, 'bed']))
    lockfil = os.path.join(tmpdir, '.'.join([prefix, 'lock']))
    if os.path.exists(outbed) or os.path.exists(lockfil):
        while os.path.exists(lockfil):
            time.sleep(0.5)
        return outbed
    else:
        lock = open(lockfil, 'wb') # aquire the file lock
        lock.close()
        digest_command = ['runHiC-digest', '-O', outbed, '-C', chromsizes_file, '--fasta-path', genomepath, '--enzyme', enzyme]
        subprocess.check_call(' '.join(digest_command), shell=True)
        os.remove(lockfil)

        return outbed

def biorep_level(pair_paths, outpre, frag_path, tmpdir):

    # a temporary file to store unfiltered all alignments
    out_total = outpre + '.total.pairsam.gz'
    merge_pairs(pair_paths, out_total, tmpdir)
    stats = collect_stats(pair_paths)['pseudo']

    # Final biorep level pairsam
    outpath = outpre + '.pairsam.gz' # select.dedup.filter

    pipeline = []
    try:
        dedup_command = ['pairtools', 'dedup', '--max-mismatch', '1', '--method', 'max', '-o', outpath, out_total]
        pipeline.append(
            subprocess.Popen(dedup_command,
                stdout=None,
                bufsize=-1)
        )
        pipeline[-1].wait()
    finally:
        sleep()
        for process in pipeline:
            if process.poll() is None:
                process.terminate()
    
    os.remove(out_total)
    
    refkey = {'cis':'410_IntraChromosomalReads',
              'trans':'420_InterChromosomalReads',
              'cis_20kb+':'412_IntraLongRangeReads(>=20Kb)',
              'total_nodups':'total_nodups'}

    substats = stats_pairs(outpath, refkey, matchpre=['dist_freq'])
    stats['130_DuplicateRemoved'] = stats['110_AfterFilteringReads'] - substats['total_nodups']
    stats['110_AfterFilteringReads'] = substats['total_nodups']
    stats['400_TotalContacts'] = stats['110_AfterFilteringReads']
    stats.update(substats)
    stats['412_IntraShortRangeReads(<20Kb)'] = stats['410_IntraChromosomalReads'] - stats['412_IntraLongRangeReads(>=20Kb)']
    del stats['total_nodups']
    
    return stats, outpath

def merge_stats(stats_pool, keys, outkey, sample_size=100000):

    stats_pool[outkey] = deepcopy(stats_pool[keys[0]])
    for i in stats_pool[outkey].keys():
        for k in keys[1:]:
            if not i in stats_pool[k]:
                continue
            if not i in ['libsize']:
                stats_pool[outkey][i] += stats_pool[k][i]
            else:
                stats_pool[outkey][i] = np.r_[stats_pool[outkey][i], stats_pool[k][i]]
    
    np.random.shuffle(stats_pool[outkey]['libsize'])
    stats_pool[outkey]['libsize'] = stats_pool[outkey]['libsize'][:sample_size] # limit sample size
    
def enzyme_level(pair_paths, outpre, keys, outkey, stats_pool, tmpdir):

    ## pair_paths --> outpre
    ## keys --> outkey
    outall = outpre + '.pairsam.gz'
    merge_pairs(pair_paths, outall, tmpdir)
    merge_stats(stats_pool, keys, outkey)

    return stats_pool, outall

def split_pairsam(pairsam_path):

    from pairtools import _fileio, _headerops
    
    # check for SAM information with the header of .pairsam.gz
    instream = _fileio.auto_open(pairsam_path, mode='r')
    header, _ = _headerops.get_header(instream)
    SAM = False
    for r in header:
        if not r.startswith('#columns:'):
            continue
        columns = r.split(':')[1].strip().split()
        if ('sam1' in columns) and ('sam2' in columns):
            SAM = True
    
    instream.close()
    pairpath = pairsam_path.replace('.pairsam.gz', '.pairs.gz')
    if SAM:
        bampath = pairsam_path.replace('.pairsam.gz', '.bam')
        split_command = ['pairtools', 'split', '--output-pairs', pairpath,
                         '--output-sam', bampath, pairsam_path]
        subprocess.check_call(' '.join(split_command), shell=True)
    else:
        mv_command = ['cp', pairsam_path, pairpath]
        subprocess.check_call(' '.join(mv_command), shell=True)
    
    # generate pairix index
    pairix_command = ['pairix', pairpath]
    subprocess.check_call(' '.join(pairix_command), shell=True)

    return pairpath






# Created on Thu Sep 13 10:44:19 2018

# Author: XiaoTao Wang

import os, subprocess, logging
from runHiC.utilities import chromsizes_from_pairs

log = logging.getLogger(__name__)

def binning_from_pairs(pairpath, res, outpath, ignore_diags=1, nproc=1, mad_max=5, min_nnz=10,
                       min_count=0, max_split=2):

    chromsizes_file, assembly = chromsizes_from_pairs(pairpath)
    
    log.log(21, 'Building contact matrix ...')
    bin_label = ':'.join([chromsizes_file, str(res)])
    bin_command = ['cooler', 'cload', 'pairix', '--assembly', assembly, '--nproc', str(nproc),
                   '--max-split', str(max_split), bin_label, pairpath, outpath]
    subprocess.check_call(' '.join(bin_command), shell=True)
    log.log(21, 'Done')

    log.log(21, 'Perform ICE ...')
    ice_command = ['cooler', 'balance', '--nproc', str(nproc), '--ignore-diags', str(ignore_diags),
                   '--mad-max', str(mad_max), '--min-nnz', str(min_nnz), '--min-count', str(min_count),
                   outpath]
    subprocess.check_call(' '.join(ice_command), shell=True)
    log.log(21, 'Done')






# Created on Fri Sep 7 19:15:32 2018

# Author: XiaoTao Wang

import os, subprocess, atexit
from runHiC.utilities import cleanFile, sleep
from runHiC.filtering import create_frag, stats_pairs, stats_samfrag
from runHiC.quality import outStatsCache

def commandExists(command):
    "Check if the bash command exists"
    command = command.split()[0]
    if subprocess.call(['which', command]) != 0:
        return False
    return True

def gzipWriter(filename):
    """
    Create a writing process with gzip or parallel gzip (pigz) attached to
    it.
    
    """
    filename = os.path.abspath(filename)
    
    with open(filename, 'wb') as outFile:
        if commandExists("pigz"):
            writer = ["pigz", "-c", "-4"]
        else:
            writer = ["gzip", "-c", "-1"]

        pwrite = subprocess.Popen(writer, stdin = subprocess.PIPE,
                                  stdout = outFile, shell = False,
                                  bufsize = -1)
    return pwrite

def uncompressSRA(filename, folder):
    
    inFile = os.path.abspath(filename)
    
    pread = subprocess.Popen(['fastq-dump', inFile, "-Z", "--split-files"],
                              stdout = subprocess.PIPE, bufsize = -1)
    
    inStream = pread.stdout
    
    outFile = os.path.split(filename)[1].replace('.sra', '') + '_{0}.fastq.gz'
    outProc1 = gzipWriter(os.path.join(folder, outFile).format(1))
    outProc2 = gzipWriter(os.path.join(folder, outFile).format(2))
    outStream1 = outProc1.stdin
    outStream2 = outProc2.stdin
    
    count = 0
    while True:
        line = inStream.readline()
        textline = line.decode()
        
        try:
            assert textline[0] == '@'
        except AssertionError:
            raise IOError('{0} is an invalid fastq file'.format(filename))
        except IndexError:
            break
        
        fastq_entry_1 = (line, inStream.readline(),
                         inStream.readline(), inStream.readline())
        outStream1.writelines(fastq_entry_1)
        
        fastq_entry_2 = (inStream.readline(), inStream.readline(),
                         inStream.readline(), inStream.readline())
        outStream2.writelines(fastq_entry_2)
        
        count += 1
    
    outProc1.communicate()
    outProc2.communicate()

def splitSRA(filename, folder, splitBy = 4000000):

    inFile = os.path.abspath(filename)
    outFile = os.path.split(inFile)[1].replace('.sra', '') + '_chunk{0}_{1}.fastq.gz'
    pread = subprocess.Popen(['fastq-dump', inFile, "-Z", "--split-files"],
                              stdout = subprocess.PIPE, bufsize = -1)
    inStream = pread.stdout

    halted = False
    counters = []
    for counter in range(1000000):

        outProc1 = gzipWriter(os.path.join(folder, outFile).format(counter, 1))
        outProc2 = gzipWriter(os.path.join(folder, outFile).format(counter, 2))
        outStream1 = outProc1.stdin
        outStream2 = outProc2.stdin

        for j in range(splitBy):

            line = inStream.readline()
            textline = line.decode()

            try:
                assert textline[0] == '@'
            except AssertionError:
                raise IOError('{0} is an invalid fastq file'.format(filename))
            except IndexError:
                halted = True
                counters.append(j)
                break

            fastq_entry = (line, inStream.readline(),
                           inStream.readline(), inStream.readline())

            outStream1.writelines(fastq_entry)
            outStream2.writelines((inStream.readline(), inStream.readline(),
                       inStream.readline(), inStream.readline()))

        outProc1.communicate()
        outProc2.communicate()
        if halted:
            return counters
        counters.append(splitBy)
    return counters


def splitSingleFastq(filename, folder, splitBy = 4000000):

    inFile = os.path.abspath(filename)
    
    parse = os.path.split(inFile)[1].split('.')[0].split('_')
    outFile = '_'.join(parse[:-1]) + '_chunk{0}_{1}.fastq.gz'
    
    if inFile.endswith('.fastq.gz'):
        pread = subprocess.Popen(['gunzip', inFile, '-c'],
                                  stdout = subprocess.PIPE, bufsize = -1)
    else:
        pread = subprocess.Popen(['cat', inFile],
                                  stdout = subprocess.PIPE, bufsize = -1)
                                  
    inStream = pread.stdout

    halted = False
    counters = []
    for counter in range(1000000):

        outProc1 = gzipWriter(os.path.join(folder, outFile).format(counter, parse[-1]))
        outStream1 = outProc1.stdin

        for j in range(splitBy):

            line = inStream.readline()
            textline = line.decode()

            try:
                assert textline[0] == '@'
            except AssertionError:
                raise IOError('{0} is not a fastq file'.format(filename))
            except IndexError:
                halted = True
                counters.append(j)
                break

            fastq_entry = (line, inStream.readline(), inStream.readline(),
                           inStream.readline())

            outStream1.writelines(fastq_entry)

        outProc1.communicate()
        counters.append(splitBy)
        
        if halted:
            return counters
        counters.append(splitBy)
        
    return counters

def buildMapIndex(aligner, genomeFolder, genomeName):
    """
    Build bwa/minimap2 index files.
    """
    lockFile = os.path.join(genomeFolder, '.'.join([genomeName, aligner, 'lock']))
    # Aquire lock
    lock = open(lockFile, 'wb')
    lock.close()

    atexit.register(cleanFile, lockFile)

    wholeGenome = os.path.join(genomeFolder, '.'.join([genomeName, 'fa']))

    if aligner=='minimap2':
        indexOut = os.path.join(genomeFolder, '.'.join([genomeName, 'mmi']))
        build_command = ['minimap2', '-x', 'sr', '-d', indexOut, wholeGenome]
    else:
        build_command = ['bwa', 'index', wholeGenome]
    
    subprocess.check_call(' '.join(build_command), shell=True)
    

def map_core(fastq_1, fastq_2, ref, outdir, aligner='minimap2', outformat='SAM', 
             nthread=1):

    outformat = '.' + outformat.lower()
    # output file name
    if fastq_1.endswith('_1.fastq.gz'):
        outpath = os.path.join(outdir,
                        os.path.split(fastq_1)[1].replace('_1.fastq.gz',outformat))
    else:
        outpath = os.path.join(outdir,
                        os.path.split(fastq_1)[1].replace('_1.fastq',outformat))
    
    # ref: reference genome index
    if aligner=='minimap2':
        map_command = ['minimap2', '-ax', 'sr', '-t', str(nthread), ref, fastq_1, fastq_2]
    else:
        map_command = ['bwa', 'mem', '-SP', '-t', str(nthread), ref, fastq_1, fastq_2]
    
    if outformat=='.sam':
        bam_command = []
    else:
        bam_command = ['samtools', 'view', '-bS', '-']
    
    pipeline = []
    try:
        # Mapping
        pipeline.append(
                subprocess.Popen(map_command,
                    stdout=subprocess.PIPE if bam_command else open(outpath, 'wb'),
                    bufsize=-1))
        
        if bam_command:
            pipeline.append(
                    subprocess.Popen(bam_command,
                        stdin=pipeline[-1].stdout,
                        stdout=open(outpath, 'wb'),
                        bufsize=-1))
        pipeline[-1].wait()
    finally:
        sleep()
        for process in pipeline:
            if process.poll() is None:
                process.terminate()
    
    return outpath

def parse_bam(bam, outfile, genomepath, chromsizes, assembly, min_mapq, max_molecule_size, max_inter_align_gap,
              walks_policy, include_readid, include_sam, drop_seq, tmpdir, enzyme):
    
    frag_path = create_frag(genomepath, chromsizes, enzyme, tmpdir)
    out_total = outfile.replace('.pairsam.gz', '.total.pairsam.gz')
    
    basic_command = ['pairtools', 'parse', '-c', chromsizes, '--assembly', assembly,
                     '--min-mapq', str(min_mapq), '--max-molecule-size', str(max_molecule_size),
                     '--max-inter-align-gap', str(max_inter_align_gap), '--walks-policy', walks_policy]
    if not include_readid:
        basic_command.append('--drop-readid')
    
    if not include_sam:
        basic_command.append('--drop-sam')
    
    if drop_seq:
        basic_command.append('--drop-seq')
    basic_command.append(bam)
    
    pipeline = []
    try:
        pipeline.append(
            subprocess.Popen(basic_command,
                stdout=subprocess.PIPE,
                bufsize=-1)
        )

        sort_command = ['pairtools', 'sort', '-o', out_total, '--nproc', '8', '--memory', '2G', '--tmpdir', tmpdir]
        pipeline.append(
            subprocess.Popen(sort_command,
                stdin=pipeline[-1].stdout,
                stdout=None,
                bufsize=-1)
        )

        pipeline[-1].wait()

    finally:
        sleep()
        for process in pipeline:
            if process.poll() is None:
                process.terminate()
    
    # stats at the bottom level
    refkey = {'total':'000_SequencedReads',
              'total_mapped':'010_DoubleSideMappedReads',
              'total_single_sided_mapped':'020_SingleSideMappedReads',
              'total_unmapped':'030_UnmappedReads'
              }
    stats = stats_pairs(out_total, refkey)
    stats['100_NormalPairs'] = stats['010_DoubleSideMappedReads']

    outpath_1 = outfile.replace('.pairsam.gz', '.select.pairsam.gz')
    pipeline = []
    try:
        select_command = ['pairtools', 'select', '(pair_type=="UU") or (pair_type=="UR") or (pair_type=="RU")',
                          '-o', outpath_1, out_total]
        pipeline.append(
            subprocess.Popen(select_command,
                stdout=None,
                bufsize=-1
            )
        )

        pipeline[-1].wait()
    
    finally:
        sleep()
        for process in pipeline:
            if process.poll() is None:
                process.terminate()
    
    os.remove(out_total)

    outpath_2 = outfile.replace('.pairsam.gz', '.select.samefrag.pairsam.gz')
    pipeline = []
    try:
        # assign fragment
        restrict_command = ['pairtools', 'restrict', '-f', frag_path, outpath_1]
        pipeline.append(
            subprocess.Popen(restrict_command,
                stdout=subprocess.PIPE,
                bufsize=-1)
        )

        ####### COLS[-6]==COLS[-3], the index may change to follow pairtools
        select_command = ['pairtools', 'select', '--output-rest', outfile, '-o', outpath_2,
                          '(COLS[-6]==COLS[-3]) and (chrom1==chrom2)']
        pipeline.append(
            subprocess.Popen(select_command,
                stdin=pipeline[-1].stdout,
                stdout=None,
                bufsize=-1)
        )

        pipeline[-1].wait()
    finally:
        sleep()
        for process in pipeline:
            if process.poll() is None:
                process.terminate()
    
    os.remove(outpath_1)

    substats, libsize = stats_samfrag(outpath_2)
    stats['110_AfterFilteringReads'] = stats['100_NormalPairs'] - substats['120_SameFragmentReads']
    stats['400_TotalContacts'] = stats['110_AfterFilteringReads']
    stats.update(substats)

    stats['libsize'] = libsize

    stats_pool = {'pseudo': stats}
    stats_pre = outfile.replace('.pairsam.gz', '.pstats') # pickled stats
    outStatsCache(stats_pool, stats_pre)




        





# Created on Thu Sep 13 21:44:24 2018
# Author: XiaoTao Wang

import os, time
import numpy as np
import matplotlib, pickle, glob
matplotlib.use('Agg')
import matplotlib.pyplot as plt

## Plot Settings
import matplotlib.pyplot as plt
# Matplotlib Settings
matplotlib.rcParams['xtick.direction'] = 'out'
matplotlib.rcParams['ytick.direction'] = 'out'
matplotlib.rcParams['axes.labelsize'] = 13
matplotlib.rcParams['xtick.labelsize'] = 13
matplotlib.rcParams['ytick.labelsize'] = 13
matplotlib.rcParams['xtick.major.size'] = 8
matplotlib.rcParams['ytick.major.size'] = 8
matplotlib.rcParams['xtick.minor.size'] = 5
matplotlib.rcParams['ytick.minor.size'] = 5
matplotlib.rcParams['xtick.major.pad'] = 6
matplotlib.rcParams['ytick.major.pad'] = 6
matplotlib.rcParams['xtick.minor.pad'] = 6
matplotlib.rcParams['ytick.minor.pad'] = 6

colorPool = ['#A6CEE3', '#1F78B4', '#B2DF8A', '#33A02C', '#348ABD', '#A60628']

def properU(pos):
    
    m_part = int(pos) // 1000000 # Megabase
    k_part = (int(pos) % 1000000) // 1000 # Kilobase
    b_part = int(pos) % 1000 # base pair
    
    if (m_part > 0) and (k_part > 0) and (b_part > 0):
        return ''.join([str(m_part), 'M', str(k_part), 'K', str(b_part), 'bp'])
    elif (m_part > 0) and (k_part > 0) and (b_part == 0):
        return ''.join([str(m_part), 'M', str(k_part), 'K'])
    elif (m_part > 0) and (k_part == 0) and (b_part > 0):
        return ''.join([str(m_part), 'M', str(b_part), 'bp'])
    elif (m_part > 0) and (k_part == 0) and (b_part == 0):
        return ''.join([str(m_part), 'M'])
    elif (m_part == 0) and (k_part > 0) and (b_part > 0):
        return ''.join([str(k_part), 'K', str(b_part), 'bp'])
    elif (m_part == 0) and (k_part > 0) and (b_part == 0):
        return ''.join([str(k_part), 'K'])
    elif (m_part == 0) and (k_part == 0) and (b_part > 0):
        return ''.join([str(b_part), 'bp'])

def printStats(stats, saveTo):

    keys = [k for k in stats if (not k in ['libsize','danglingStart']) and (not k.startswith('dist_freq'))]

    longrange = stats['412_IntraLongRangeReads(>=20Kb)']
    contacts = stats['400_TotalContacts']
    
    Total = stats['000_SequencedReads']
    DUreads = stats['010_DoubleSideMappedReads']
    selfLig = stats['122_SelfLigationReads']
    dangling = stats['124_DanglingReads']
    Uratio = float(DUreads) / Total
    Fratio = float(selfLig) / Total
    Dratio = float(dangling) / Total
    usage = float(contacts) / Total
    longRatio = float(longrange) / contacts

    # Tree-like print
    with open(saveTo, 'w') as myfile:
        for i in sorted(keys):
            if (i[2] != '0'):
                myfile.write('\t\t')
            elif (i[1] != '0') and (i[2] == '0'):
                myfile.write('\t')
            myfile.write(str(i))
            myfile.write(':  ')
            myfile.write(str(stats[i]))
            myfile.write('\n')
        myfile.write('\nCritical Indicators:\n')
        myfile.write('Double Unique Mapped Ratio = {0} / {1} = {2:.4f}\n'.format(DUreads, Total, Uratio))
        myfile.write('Self-Ligation Ratio = {0} / {1} = {2:.4f}\n'.format(selfLig, Total, Fratio))
        myfile.write('Dangling-Reads Ratio = {0} / {1} = {2:.4f}\n'.format(dangling, Total, Dratio))
        myfile.write('Long-Range Ratio = {0} / {1} = {2:.4f}\n'.format(longrange, contacts, longRatio))
        myfile.write('Data Usage = {0} / {1} = {2:.4f}\n'.format(contacts, Total, usage))


def typePlot(stats, outfile, dpi = 300):

    from collections import defaultdict

    LT = {}; RT = {}; IT = {}; OT = {}
    Total = defaultdict(int)
    for k in stats:
        if not k.startswith('dist_freq'):
            continue
        _, dist, t = k.split('/')
        if not '-' in dist:
            continue
        dist = int(dist.split('-')[1])
        if (dist > 10000000) and (dist <= 1):
            continue
        if t == '++':
            RT[dist] = stats[k]
        elif t == '--':
            LT[dist] = stats[k]
        elif t == '+-':
            IT[dist] = stats[k]
        else:
            OT[dist] = stats[k]

        Total[dist] += stats[k]
    
    lt = []; rt = []; it = []; ot = []
    for k in sorted(LT):
        if Total[k] > 0:
            lt.append(LT[k]/Total[k])
            rt.append(RT[k]/Total[k])
            it.append(IT[k]/Total[k])
            ot.append(OT[k]/Total[k])
        else:
            lt.append(np.nan)
            rt.append(np.nan)
            it.append(np.nan)
            ot.append(np.nan)
    lt = np.r_[lt]; rt = np.r_[rt]; it = np.r_[it]; ot = np.r_[ot]

    x = np.arange(0, 8, 0.25)
    xticks = list(range(1, 8))
    xticklabels = ['10bp', '100bp', '1K', '10K', '100K', '1M', '10M']

    fig = plt.figure(figsize = (15, 9))
    ax = fig.add_subplot(111)
    lines = []
    labels = ['Left Type', 'Right Type', 'Inner Type', 'Outer Type']
    idx = 0
    for y in [lt, rt, it, ot]:
        L = ax.plot(x, y[:x.size], color = colorPool[idx], linewidth = 2)
        lines.extend(L)
        idx += 1
    
    ax.set_xticks(xticks)
    ax.set_xticklabels(xticklabels)
    ax.set_xlabel('Genomic Separation')
    ax.set_ylabel('Type Ratio')
    
    ax.set_ylim((0, 1))
    ax.set_title('Read Pair Type Statistics')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
        
    ax.legend(lines, labels, frameon = False, fontsize = 11, labelspacing = 1,
            handletextpad = 1, borderpad = 1, markerscale = 1, numpoints = 1,
            ncol = 2, loc = 'upper right')
        
    plt.savefig(outfile, dpi = dpi)
    plt.close()
    
def plot_libsize(stats, outplot, dpi = 300):

    libsize = stats['libsize']

    low = np.percentile(libsize, 0.2)
    high = np.percentile(libsize, 99.8)
    libsize = libsize[(libsize<=high) & (libsize>=low)]
    fig = plt.figure(figsize = (15, 9))
    ax = fig.add_subplot(111)
    ax.hist(libsize, bins = 15, color = colorPool[-1])
    ax.set_title('Estimated Library Size Distribution (bp)')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    plt.savefig(outplot, dpi = dpi)
    plt.close()

def plot_dangling_details(stats, outplot, dpi = 300):

    danglingStart = stats['danglingStart']

    danglingStart = danglingStart[danglingStart<=0.5]
    fig = plt.figure(figsize = (15, 9))
    ax = fig.add_subplot(111)
    ax.hist(danglingStart, bins = 15, color = colorPool[-2])
    ax.set_title('Start Site of Dandling Ends Relative to Fragments (Ratio)')
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    plt.savefig(outplot, dpi = dpi)
    plt.close()


def outStatsCache(stats_pool, outpre):

    lockfil = outpre + '.lock'
    while os.path.exists(lockfil):
        time.sleep(0.5)
    
    lock = open(lockfil, 'wb') # aquire the file lock
    lock.close()

    pattern = outpre + '*'
    allfile = glob.glob(pattern)
    allfile.remove(lockfil)
    if not len(allfile):
        cache = outpre + '.1'
    else:
        counts = [int(f.split('.')[-1]) for f in allfile]
        counts.sort()
        suffix = str(counts[-1]+1)
        cache = '.'.join([outpre, suffix])
    
    with open(cache, 'wb') as out:
        pickle.dump(stats_pool, out)
    
    os.remove(lockfil) # release the lock


def loadStats(cache_pre):

    pattern = cache_pre + '*'
    allfile = glob.glob(pattern)
    if not len(allfile):
        raise Exception('Stats cache {0} cannot be found, exit'.format(cache_pre))
    stats_pool = {}
    for f in allfile:
        with open(f, 'rb') as source:
            tmp = pickle.load(source)
        stats_pool.update(tmp)
    
    return stats_pool

def checkKeys(stats_pool, keys):

    return all([(k in stats_pool) for k in keys])

def update_stats_pool(stats_pool, keys, cache_pre):

    pattern = cache_pre + '*'
    allfile = glob.glob(pattern)
    for f in allfile:
        with open(f, 'rb') as source:
            tmp = pickle.load(source)
        for k in keys:
            if k in tmp:
                stats_pool[k] = tmp[k]
                if os.path.exists(f):
                    os.remove(f) # clean duplicate records

    return stats_pool



# Created on Wed Dec 24 15:26:26 2014

# Author: XiaoTao Wang

"""
Setup script for hicpeaks.

This is a free software under GPLv3. Therefore, you can modify, redistribute
or even mix it with other GPL-compatible codes. See the file LICENSE
included with the distribution for more details.

"""
import os, sys, runHiC, glob
import setuptools

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if (sys.version_info.major!=3) or (sys.version_info.minor<5):
    print('PYTHON 3.5+ IS REQUIRED. YOU ARE CURRENTLY USING PYTHON {}'.format(sys.version.split()[0]))
    sys.exit(2)

# Guarantee Unix Format
for src in glob.glob('scripts/*'):
    text = open(src, 'r').read().replace('\r\n', '\n')
    open(src, 'w').write(text)

setuptools.setup(
    name = 'runHiC',
    version = runHiC.__version__,
    author = runHiC.__author__,
    author_email = 'wangxiaotao686@gmail.com',
    url = 'https://github.com/XiaoTaoWang/HiC_pipeline',
    description = 'A easy-to-use Hi-C processing software supporting distributed computation',
    keywords = 'Hi-C Arima ICE cooler pairs bioinformatics pipeline',
    long_description = read('README.rst'),
    long_description_content_type='text/x-rst',
    scripts = glob.glob('scripts/*'),
    packages = setuptools.find_packages(),
    classifiers = [
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        ]
    )
